@@ -5,10 +5,14 @@
 //! the `sys.monitoring.DISABLE` sentinel.
 
 pub mod code_object;
+mod errors;
+mod ffi;
 mod logging;
 pub mod monitoring;
+mod policy;
 mod runtime;
 mod session;
+pub mod trace_filter;
 
 pub use crate::code_object::{CodeObjectRegistry, CodeObjectWrapper};
 pub use crate::monitoring as tracer;
@@ -24,11 +28,15 @@ use pyo3::prelude::*;
 #[pymodule]
 fn codetracer_python_recorder(_py: Python<'_>, m: &Bound<'_, PyModule>) -> PyResult<()> {
     // Initialize logging on import so users see logs without extra setup.
-    // Respect RUST_LOG if present; otherwise default to debug for this crate.
-    logging::init_rust_logging_with_default("codetracer_python_recorder=debug");
+    // Respect RUST_LOG if present; otherwise keep our crate quiet unless warnings/errors occur.
+    logging::init_rust_logging_with_default("codetracer_python_recorder=warn");
+    ffi::register_exceptions(m)?;
     m.add_function(wrap_pyfunction!(start_tracing, m)?)?;
     m.add_function(wrap_pyfunction!(stop_tracing, m)?)?;
     m.add_function(wrap_pyfunction!(is_tracing, m)?)?;
     m.add_function(wrap_pyfunction!(flush_tracing, m)?)?;
+    m.add_function(wrap_pyfunction!(policy::configure_policy_py, m)?)?;
+    m.add_function(wrap_pyfunction!(policy::py_configure_policy_from_env, m)?)?;
+    m.add_function(wrap_pyfunction!(policy::py_policy_snapshot, m)?)?;
     Ok(())
 }
